import lexer;
import parser;
import graph;
import std.stdio;
import std.string;
import std.ctype;
import std.conv : to;

//////////////////////////////////////////////////////////////////////////////////////
// ソースを字句解析して構文解析してコンパイルして返す
//////////////////////////////////////////////////////////////////////////////////////

GNode compile( string src )
{
	return compile( parse(src) )["main"];
}

//////////////////////////////////////////////////////////////////////////////////////

class CompileFail : Exception
{
	this(string msg, PEntity pe){super(msg ~ " @ " ~ to!(string)(pe.line));}
}

class RuntimeFail : Exception
{
	this(string msg, PEntity pe=null){super(msg ~ " @ " ~ (pe ? to!(string)(pe.line) : "?"));}
}

Env dup(Env e)
{
	Env e2; foreach(a,b;e) e2[a]=b;
	return e2;
}

//////////////////////////////////////////////////////////////////////////////////////
// Value の定義
//////////////////////////////////////////////////////////////////////////////////////

class VInt : Value
{
	long v;

	this(long vv){v = vv;}
	string toString() { return to!(string)(v); }
}

long AS_INT( Value v, PEntity pe )
{
	if( auto v = cast(VInt)v )
		return v.v;
	throw new RuntimeFail( to!(string)(v) ~ " is not an integer", pe );
}

class VLambda : Value
{
	Env    e;
	string par;
	PExpr  bod;

	GNode apply( GNode arg )
	{
		Env e2 = e.dup();
		e2[par] = arg;
		return compile( bod, e2 );
	}

	this(Env ee, string pp, PExpr bb){e=ee; par=pp; bod=bb;}
	string toString() { return par ~ " -> " ~ to!(string)(bod); }
}

VLambda AS_LAMBDA( Value v, PEntity pe )
{
	if( auto v = cast(VLambda)v )
		return v;
	throw new RuntimeFail( to!(string)(v) ~ " is not a function", pe );
}

class VTag : Value
{
	string  tag;
	GNode[] mem;

	this(string tt, GNode[] mm){tag=tt; mem=mm;}
	string toString() { return tag ~ (mem.length==0 ? "" : to!(string)(mem)); }
}

VTag AS_TAG( Value v, string tagName, PEntity pe=null )
{
	if( auto v = cast(VTag)v )
		return v.tag==tagName ? v : null;
	throw new RuntimeFail( to!(string)(v) ~ " is not a tag value", pe );
}

Value int_(T)( T v )
{
	return new VInt( to!(long)(v) );
}

Value bool_(T)( T v )
{
	return tag_( v?"True":"False" );
}

Value tag_( string t, GNode[] g... )
{
	return new VTag( t, g.dup );
}

//////////////////////////////////////////////////////////////////////////////////////
// 構文木をグラフに中途半端にコンパイル
//////////////////////////////////////////////////////////////////////////////////////

GNode compile( PExpr pe, Env env )
{
	if( auto pe = cast(PBinOp)pe )
	{
		// 二項演算子
		GNode ga = compile(pe.a, env);
		GNode gb = compile(pe.b, env);
		switch(pe.op)
		{
		case "+":   return new GNode({return int_( AS_INT(ga.run,pe) + AS_INT(gb.run,pe) );});
		case "-":   return new GNode({return int_( AS_INT(ga.run,pe) - AS_INT(gb.run,pe) );});
		case "*":   return new GNode({return int_( AS_INT(ga.run,pe) * AS_INT(gb.run,pe) );});
		case "/":   return new GNode({return int_( AS_INT(ga.run,pe) / AS_INT(gb.run,pe) );});
		case "%":   return new GNode({return int_( AS_INT(ga.run,pe) % AS_INT(gb.run,pe) );});
		case "<":   return new GNode({return bool_( AS_INT(ga.run,pe) < AS_INT(gb.run,pe) );});
		case ">":   return new GNode({return bool_( AS_INT(ga.run,pe) > AS_INT(gb.run,pe) );});
		case "==":  return new GNode({return bool_( AS_INT(ga.run,pe) == AS_INT(gb.run,pe) );});
		case "!=":  return new GNode({return bool_( AS_INT(ga.run,pe) != AS_INT(gb.run,pe) );});
		case ">=":  return new GNode({return bool_( AS_INT(ga.run,pe) >= AS_INT(gb.run,pe) );});
		case "<=":  return new GNode({return bool_( AS_INT(ga.run,pe) <= AS_INT(gb.run,pe) );});
		case ">>":  return new GNode({return tag_("SEQIO", ga, gb);});
		}
	}
	else
	if( auto pe = cast(PAppl)pe )
	{
		// 関数適用
		GNode ga = compile(pe.a, env);
		GNode gb = compile(pe.b, env);
		return new GNode({return AS_LAMBDA(ga.run,pe).apply(gb).run;});
	}
	else
	if( auto pe = cast(PFun)pe )
	{
		// λ
		return new GNode(new VLambda(env, pe.p, pe.bod));
	}
	else
	if( auto pe = cast(PCase)pe )
	{
		// case式
		GNode gc = compile(pe.c, env);
		return new GNode({
			foreach(i,pat; pe.ps)
				if( Env e1 = pattern_match(pat, gc) )
				{
					Env e2 = env.dup();
					foreach(k,v; e1) e2[k] = v;
					return compile( pe.es[i], e2 ).run;
				}
			throw new RuntimeFail("no matching case", pe);
		});
	}
	else
	if( auto pe = cast(PTag)pe )
	{
		// タグ付き値
		GNode[] gs;
		foreach(e; pe.es) gs ~= compile(e, env);
		return new GNode(new VTag(pe.name, gs));
	}
	else
	if( auto pe = cast(PVar)pe )
	{
		// 変数
		string v = pe.v;
		if( v!="f" && isNumeric(v) )
			return new GNode(int_(v));
		else if( v in env )
			return env[v];
		throw new CompileFail("variable "~v~" not defined", pe);
	}
}

Env compile( PProgram pp )
{
	// 再帰定義をちゃんと扱うための大変てぬきなトリック
	Env env;
	foreach(dd; pp.decl) env[dd.name] = new GNode;
	foreach(dd; pp.decl) env[dd.name].set( compile(dd.expr, env) );
	return env;
}

//////////////////////////////////////////////////////////////////////////////////////
// パターンマッチ。再帰的に。マッチしたら変数束縛を返す
//////////////////////////////////////////////////////////////////////////////////////

Env pattern_match( Pat p, GNode g )
{
	Env e = null;
	if(auto p = cast(VarPat)p)
	{
		e[p.v] = g;
	}
	else if(auto p = cast(TagPat)p)
	{
		if(auto t = AS_TAG(g.run, p.tag, p))
			if(p.vs.length == t.mem.length)
				if( p.vs.length == 0 )
					e["______________dummy_______________"] = null;
				else
					foreach(j,subpat; p.vs)
					{
						Env e2 = pattern_match(subpat, t.mem[j]);
						if( e2 is null )
							return null;
						foreach(k,v; e2) e[k] = v;
					}
	}
	else
	{
		throw new RuntimeFail("Nazo no pattern", p);
	}
	return e;
}
