import lexer;
import std.ctype;
import std.typetuple;
import std.conv : to;

//////////////////////////////////////////////////////////////////////////////////////
// 文字列渡すと字句解析して構文解析して結果を返すよ
//////////////////////////////////////////////////////////////////////////////////////

PProgram parse( string src )
{
	return new PProgram( lex(src) );
}

//////////////////////////////////////////////////////////////////////////////////////
// ほじょるーちん
//////////////////////////////////////////////////////////////////////////////////////

class ParseFail : Exception
{
	this( Token t, string msg )
	{
		super( msg ~ " @ " ~ (t is null ? "end" : to!(string)(t)) );
	}
}

void eat( string v, ref Token[] ts )
{
	if( ts.length == 0 )
		throw new ParseFail(null,  "missing "~v);
	else if( ts[0].value != v )
		throw new ParseFail(ts[0], "missing "~v);
	ts = ts[1..$];
}

string eat_one( ref Token[] ts )
{
	if( ts.length == 0 )
		throw new ParseFail(null, "missing something");
	scope(success) ts = ts[1..$];
	return ts[0].value;
}

bool nextis( Token[] ts, string[] tok... )
{
	if( ts.length == 0 )
		return false;
	foreach(t; tok)
		if( ts[0].value == t )
			return true;
	return false;
}

bool nextisnot( Token[] ts, string[] tok... )
{
	if( ts.length == 0 )
		return false;
	foreach(t; tok)
		if( ts[0].value == t )
			return false;
	return true;
}

class PEntity
{
	int line = -2;
	this() {}
	this( Token[] ts ) { line = (ts.length ? ts[0].line : -1); }
	void setpos(PEntity pe) {line = pe.line;}
}

//////////////////////////////////////////////////////////////////////////////////////
// 構文解析本体
//////////////////////////////////////////////////////////////////////////////////////

///
/// PROGRAM = (DATADECL | DECL)*
///

class PProgram : PEntity
{
	PDecl[] decl;

	this( Token[] ts )
	{
		super(ts);
		PDecl2[] decl2;
		while( ts.length )
			if( ts[0].value=="data" )
				decl ~= parse_datadecl(ts);
			else
				decl2 ~= new PDecl2(ts);
		unifydefs(decl2);
	}

	void unifydefs( PDecl2[] decl2 )
	{
		PDecl2[][string] ds;
		foreach(dd; decl2)
			ds[dd.name] ~= dd;
		foreach(name, dds; ds)
		{
			// a b c = d
			// a e f = g
			//-->
			//  a = \x0->\x1->case [Tuple x0 x1] of
			//                   { (Tuple b c) -> d ;
			//                     (Tuple e f) -> g }

			int n = dds[0].ps.length;
			foreach(d; dds)
				if( d.ps.length != n )
					throw new ParseFail(null, name~" different number of params");
			if( n == 0 )
				decl ~= new PDecl(name, dds[0].expr);
			else
			{
				PTag tup = new PTag("Tuple", n);

				Pat[] ps;
				PExpr[] es;
				foreach(d; dds)
				{
					ps ~= new TagPat("Tuple", d.ps);
					es ~= d.expr;
				}
				PExpr e = new PCase(tup, ps, es);
				e.setpos(dds[0]);
				for(int i=n-1; i>=0; --i)
					e = new PFun(i, e);
				decl ~= new PDecl(name, e);
			}
		}
	}

	string toString()
	{
		string s;
		foreach(d; decl)
			s ~= to!(string)(d) ~ "\n";
		return s;
	}
}

///
/// DATADECL = 'data' _ = TAG _* ('|' TAG _*) ';;'
///

PDecl[] parse_datadecl(ref Token[] ts)
{
	PDecl[] pd;

	eat("data", ts);
	eat_one(ts); // name
	eat("=", ts);
	while( ts.nextisnot(";;") )
	{
		string name = eat_one(ts);
		int n = 0;
		while( ts.nextisnot("|", ";;") )
			++n, eat_one(ts);

		PExpr e = new PTag(name, n);
		for(int i=n-1; i>=0; --i)
			e = new PFun(i, e);
		pd ~= new PDecl(name, e);

		if( ts.nextis("|") )
			eat_one(ts);
	}
	eat(";;", ts);

	return pd;
}

///
/// DECL = NAME PAT* '=' EXPR ';;'
///

class PDecl : PEntity
{
	string name;
	PExpr  expr;
	this( string n, PExpr pe ) {name=n; expr=pe;}

	this( ref Token[] ts )
	{
		super(ts);
		name = eat_one(ts);
		eat("=", ts);
		expr = parse_expr(ts);
		eat(";;", ts);
	}

	string toString()
	{
		return to!(string)(name) ~ " := " ~ to!(string)(expr);
	}
}

class PDecl2 : PEntity
{
	string name;
	Pat[]    ps;
	PExpr  expr;

	this( ref Token[] ts )
	{
		super(ts);
		name = eat_one(ts);
		while( ts.nextisnot("=") )
			ps ~= parse_pat(ts);
		eat("=", ts);
		expr = parse_expr(ts);
		eat(";;", ts);
	}
}

///
/// EXPR ::= APPL (BINOP APPL)*
/// APPL ::= UNIT+
///

class PExpr : PEntity
{
	this() {}
	this(Token[] ts) { super(ts); }
}

class PBinOp : PExpr
{
	string op;
	PExpr  a;
	PExpr  b;
	this( string oo, PExpr aa, PExpr bb ) {op=oo; a=aa; b=bb;}

	string toString()
	{
		return to!(string)(a) ~ " "~to!(string)(op)~" " ~ to!(string)(b);
	}
}

class PAppl : PExpr
{
	PExpr a;
	PExpr b;
	this( PExpr aa, PExpr bb ) {a=aa; b=bb;}

	string toString()
	{
		return to!(string)(a) ~ "(" ~ to!(string)(b) ~ ")";
	}
}

alias TypeTuple!("+" ,"-" ,"*" ,"/" ,"%" ,"<" ,">","==" ,"<=" ,">=" ,"!=" ,">>" ,"$") binop;

PExpr parse_expr( ref Token[] ts )
{
	PExpr a = parse_appl(ts);
	while( ts.nextis(binop) )
	{
		string op = ts[0].value;
		ts = ts[1..$];
		a = new PBinOp(op, a, parse_appl(ts));
	}
	return a;
}

PExpr parse_appl( ref Token[] ts )
{
	PExpr a = parse_unit(ts);
	while( ts.nextisnot( ")", ";", ";;", "}", "of", binop ) )
		a = new PAppl(a, parse_unit(ts));
	return a;
}

/// UNIT ::= '(' EXPR ')
///        | '[' NAME UNIT* ']'
///        | '\' NAME '-> EXPR
///        | 'case' EXPR 'of' '{' PAT '->' EXPR (';' PAT '->' EXPR)* '}'
///        | VAR

PExpr parse_unit( ref Token[] ts )
{
	if( ts.length==0 )
		throw new ParseFail(null, "premature end of input");

	switch( ts[0].value )
	{
	case "(":
		eat("(", ts);
		scope(success) eat(")", ts);
		return parse_expr(ts);
	case "\\":   return new PFun(ts);
	case "[":    return new PTag(ts);
	case "case": return new PCase(ts);
	default:     return new PVar(ts);
	}
}

class PTag : PExpr
{
	string name;
	PExpr[] es;

	this( string name, int n ) // てぬき君
	{
		this.name = name;
		foreach(i; 0..n)
			es ~= new PVar("x"~to!(string)(i));
	}

	this( ref Token[] ts )
	{
		super(ts);
		eat("[", ts);
		name = eat_one(ts);
		while( ts.length && ts[0].value!="]" )
			es ~= parse_unit(ts);
		eat("]", ts);
	}
	string toString()
	{
		return name~to!(string)(es);
	}
}


class PFun : PExpr
{
	this( int i, PExpr b ) // てぬきくん
	{
		p = "x"~to!(string)(i);
		bod = b;
	}

	this( ref Token[] ts )
	{
		super(ts);
		eat("\\", ts);
		p = eat_one(ts);
		eat("->", ts);
		bod = parse_expr(ts);
	}
	string p;
	PExpr  bod;

	string toString()
	{
		return to!(string)(p) ~ " -> " ~ to!(string)(bod);
	}
}

class PCase : PExpr
{
	PExpr   c;
	Pat[]   ps;
	PExpr[] es;
	this( PExpr cc, Pat[] pp, PExpr[] ee ) {c=cc;ps=pp;es=ee;}
	this( ref Token[] ts )
	{
		super(ts);
		eat("case", ts);
		c = parse_expr(ts);
		eat("of", ts);
		eat("{", ts);

		ps ~= parse_pat(ts);
		eat("->", ts);
		es ~= parse_expr(ts);
		while( ts.nextis(";") )
		{
			eat(";", ts);
			ps ~= parse_pat(ts);
			eat("->", ts);
			es ~= parse_expr(ts);
		}
		eat("}", ts);
	}
	string toString()
	{
		return "case(" ~ to!(string)(c) ~ ")of"~to!(string)(ps)~" -> "~to!(string)(es)~"}";
	}
}

class PVar : PExpr
{
	string v;

	this( string v ) { this.v = v; }
	this( ref Token[] ts )
	{
		super(ts);
		v = eat_one(ts);
	}

	string toString()
	{
		return v;
	}
}

///
/// PAT ::= VAR | '(' TAG PAT* ')'
///

class Pat : PEntity
{
	this(){}
	this( ref Token[] ts ) { super(ts); }
}

Pat parse_pat( ref Token[] ts )
{
	if( ts.length && ts[0].value=="(" )
		return new TagPat(ts,true);
	if( ts.length && isupper(ts[0].value[0]) )
		return new TagPat(ts,false);
	return new VarPat(ts);
}

class VarPat : Pat
{
	string v;
	this( ref Token[] ts )
	{
		super(ts);
		v = eat_one(ts);
	}

	string toString()
	{
		return v;
	}
}

class TagPat : Pat
{
	string tag;
	Pat[] vs;
	this(string t, Pat[] v){tag=t; vs=v;}//てぬきくん
	this( ref Token[] ts, bool kakko )
	{
		super(ts);
		if(kakko)
			eat("(", ts);
		tag = eat_one(ts);
		if(kakko)
		{
			while( ts.length && ts[0].value!=")" )
				vs ~= parse_pat(ts);
			eat(")", ts);
		}
	}

	string toString()
	{
		return "("~tag~to!(string)(vs)~")";
	}
}
